"""Import upstream source into a branch"""

import os
from shutil import rmtree
from StringIO import StringIO
import tarfile
from unittest import makeSuite

from bzrlib.bzrdir import BzrDir
from bzrlib.delta import compare_trees
from bzrlib.errors import NoSuchFile, BzrCommandError
from bzrlib.osutils import pathjoin, isdir
from bzrlib.tests import TestCaseInTempDir
from bzrlib.transform import TreeTransform
from bzrlib.workingtree import WorkingTree


def top_directory(path):
    """Return the top directory given in a path."""
    dirname = os.path.dirname(path)
    last_dirname = dirname
    while True:
        dirname = os.path.dirname(dirname)
        if dirname == '' or dirname == last_dirname:
            return last_dirname
        last_dirname = dirname


def common_directory(names):
    """Determine a single directory prefix from a list of names"""
    possible_prefix = None
    for name in names:
        name_top = top_directory(name)
        if possible_prefix is None:
            possible_prefix = name_top
        else:
            if name_top != possible_prefix:
                return None
    return possible_prefix


def import_tar(tree, tar_input):
    """Replace the contents of a working directory with tarfile contents.
    The tarfile may be a gzipped or bzipped stream.  File ids will be updated.
    """
    tar_file = tarfile.open('lala', 'r', tar_input)
    prefix = common_directory(tar_file.getnames())
    added = []
    tt = TreeTransform(tree)
    for path, entry in tree.inventory.iter_entries():
        trans_id = tt.trans_id_tree_path(path)
        tt.delete_contents(trans_id)

    for member in tar_file.getmembers():
        relative_path = member.name 
        if prefix is not None:
            relative_path = relative_path[len(prefix)+1:]
        if relative_path == '':
            continue
        trans_id = tt.trans_id_tree_path(relative_path)
        added.append(relative_path)
        path = tree.abspath(relative_path)
        if member.isreg():
            tt.create_file([tar_file.extractfile(member).read()], trans_id)
        elif member.isdir():
            if isdir(path):
                tt.cancel_deletion(trans_id)
            else:
                tt.create_directory(trans_id)
        elif member.issym():
            tt.create_symlink(member.linkname, trans_id)
    tt.apply()
    update_ids(tree, added)


def update_ids(tree, added):
    """Make sure that all present files files have file_ids.
    """
    # XXX detect renames
    for path in added:
        if tree.path2id(path) is None:
            tree.add(path)


def do_import(source):
    """Implementation of import command.  Intended for UI only"""
    tree = WorkingTree.open_containing('.')[0]
    tree.lock_write()
    try:
        if compare_trees(tree, tree.basis_tree()).has_changed():
            raise BzrCommandError("Working tree has uncommitted changes.")

        if (source.endswith('.tar') or source.endswith('.tar.gz') or 
            source.endswith('.tar.bz2')):
            try:
                tar_input = file(source, 'rb')
            except:
                if e.errno == errno.ENOENT:
                    raise NoSuchFile(source)
            try:
                import_tar(tree, tar_input)
            finally:
                tar_input.close()
    finally:
        tree.unlock()

class TestImport(TestCaseInTempDir):

    def make_tar(self):
        result = StringIO()
        tar_file = tarfile.open('project-0.1.tar', 'w', result)
        os.mkdir('project-0.1')
        tar_file.add('project-0.1')
        os.mkdir('project-0.1/junk')
        tar_file.add('project-0.1/junk')
        
        f = file('project-0.1/README', 'wb')
        f.write('What?')
        f.close()
        tar_file.add('project-0.1/README')

        f = file('project-0.1/FEEDME', 'wb')
        f.write('Hungry!!')
        f.close()
        tar_file.add('project-0.1/FEEDME')

        tar_file.close()
        rmtree('project-0.1')
        result.seek(0)
        return result

    def make_tar2(self):
        result = StringIO()
        tar_file = tarfile.open('project-0.2.tar', 'w', result)
        os.mkdir('project-0.2')
        tar_file.add('project-0.2')
        
        os.mkdir('project-0.2/junk')
        tar_file.add('project-0.2/junk')

        f = file('project-0.2/README', 'wb')
        f.write('Now?')
        f.close()
        tar_file.add('project-0.2/README')

        tar_file.add('project-0.2/README')

        tar_file.close()
        rmtree('project-0.2')
        result.seek(0)
        return result

    def make_messed_tar(self):
        result = StringIO()
        tar_file = tarfile.open('project-0.1.tar', 'w', result)
        os.mkdir('project-0.1')
        tar_file.add('project-0.1')

        os.mkdir('project-0.2')
        tar_file.add('project-0.2')
        
        f = file('project-0.1/README', 'wb')
        f.write('What?')
        f.close()
        tar_file.add('project-0.1/README')
        tar_file.close()
        rmtree('project-0.1')
        result.seek(0)
        return result

    def test_top_directory(self):
        self.assertEqual(top_directory('ab/b/c'), 'ab')
        self.assertEqual(top_directory('/etc'), '/')

    def test_common_directory(self):
        self.assertEqual(common_directory(['ab/c/d', 'ab/c/e']), 'ab')
        self.assertIs(common_directory(['ab/c/d', 'ac/c/e']), None)

    def test_untar(self):
        tar_file = self.make_tar()
        tree = BzrDir.create_standalone_workingtree('tree')
        import_tar(tree, tar_file)
        self.assertTrue(tree.path2id('README') is not None) 
        self.assertTrue(tree.path2id('FEEDME') is not None)
        self.assertTrue(os.path.isfile(tree.abspath('README')))
        self.assertEqual(tree.inventory[tree.path2id('README')].kind, 'file')
        self.assertEqual(tree.inventory[tree.path2id('FEEDME')].kind, 'file')
        
        f = file(tree.abspath('junk/food'), 'wb')
        f.write('I like food\n')
        f.close()

        tar_file = self.make_tar2()
        import_tar(tree, tar_file)
        self.assertTrue(tree.path2id('README') is not None) 
        self.assertTrue(not os.path.exists(tree.abspath('FEEDME')))


    def test_untar2(self):
        tar_file = self.make_messed_tar()
        tree = BzrDir.create_standalone_workingtree('tree')
        import_tar(tree, tar_file)
        self.assertTrue(tree.path2id('project-0.1/README') is not None) 


def test_suite():
    return makeSuite(TestImport)
